<?php

namespace Box\Spout\Reader\XLSX\Helper;

/**
 * Class DateFormatHelper
 * This class provides helper functions to format Excel dates
 */
class DateFormatHelper
{
    const KEY_GENERAL = 'general';
    const KEY_HOUR_12 = '12h';
    const KEY_HOUR_24 = '24h';

    /**
     * This map is used to replace Excel format characters by their PHP equivalent.
     * Keys should be ordered from longest to smallest.
     *
     * @var array Mapping between Excel format characters and PHP format characters
     */
    private static $excelDateFormatToPHPDateFormatMapping = [
        self::KEY_GENERAL => [
            'am/pm' => 'A',  // Uppercase Ante meridiem and Post meridiem
            ':mm'   => ':i', // Minutes with leading zeros - if preceded by a ":" (otherwise month)
            'mm:'   => 'i:', // Minutes with leading zeros - if followed by a ":" (otherwise month)
            'ss'    => 's',  // Seconds, with leading zeros
            '.s'    => '',   // Ignore (fractional seconds format does not exist in PHP)

            'e'     => 'Y',  // Full numeric representation of a year, 4 digits
            'yyyy'  => 'Y',  // Full numeric representation of a year, 4 digits
            'yy'    => 'y',  // Two digit representation of a year
            'mmmmm' => 'M',  // Short textual representation of a month, three letters ("mmmmm" should only contain the 1st letter...)
            'mmmm'  => 'F',  // Full textual representation of a month
            'mmm'   => 'M',  // Short textual representation of a month, three letters
            'mm'    => 'm',  // Numeric representation of a month, with leading zeros
            'm'     => 'n',  // Numeric representation of a month, without leading zeros
            'dddd'  => 'l',  // Full textual representation of the day of the week
            'ddd'   => 'D',  // Textual representation of a day, three letters
            'dd'    => 'd',  // Day of the month, 2 digits with leading zeros
            'd'     => 'j',  // Day of the month without leading zeros
        ],
        self::KEY_HOUR_12 => [
            'hh'    => 'h',  // 12-hour format of an hour without leading zeros
            'h'     => 'g',  // 12-hour format of an hour without leading zeros
        ],
        self::KEY_HOUR_24 => [
            'hh'    => 'H',  // 24-hour hours with leading zero
            'h'     => 'G',  // 24-hour format of an hour without leading zeros
        ],
    ];

    /**
     * Converts the given Excel date format to a format understandable by the PHP date function.
     *
     * @param string $excelDateFormat Excel date format
     * @return string PHP date format (as defined here: http://php.net/manual/en/function.date.php)
     */
    public static function toPHPDateFormat($excelDateFormat)
    {
        $dateFormat = \preg_replace('/^(?:\[\$[^\]]+?\])?([^;]*).*/', '$1', $excelDateFormat);

        $dateFormatParts = \explode('"', $dateFormat);

        foreach ($dateFormatParts as $partIndex => $dateFormatPart) {
            if ($partIndex % 2 === 1) {
                continue;
            }

            $transformedPart = \strtolower($dateFormatPart);

            $transformedPart = \str_replace('\\', '', $transformedPart);

            $transformedPart = \strtr($transformedPart, self::$excelDateFormatToPHPDateFormatMapping[self::KEY_GENERAL]);

            if (self::has12HourFormatMarker($dateFormatPart)) {
                $transformedPart = \strtr($transformedPart, self::$excelDateFormatToPHPDateFormatMapping[self::KEY_HOUR_12]);
            } else {
                $transformedPart = \strtr($transformedPart, self::$excelDateFormatToPHPDateFormatMapping[self::KEY_HOUR_24]);
            }

            $dateFormatParts[$partIndex] = $transformedPart;
        }

        $phpDateFormat = \implode('"', $dateFormatParts);

        $phpDateFormat = \preg_replace_callback('/"(.+?)"/', function ($matches) {
            $stringToEscape = $matches[1];
            $letters = \preg_split('//u', $stringToEscape, -1, PREG_SPLIT_NO_EMPTY);

            return '\\' . \implode('\\', $letters);
        }, $phpDateFormat);

        return $phpDateFormat;
    }

    /**
     * @param string $excelDateFormat Date format as defined by Excel
     * @return bool Whether the given date format has the 12-hour format marker
     */
    private static function has12HourFormatMarker($excelDateFormat)
    {
        return (\stripos($excelDateFormat, 'am/pm') !== false);
    }
}
