<?php
/**
 * CodeIgniter library
 *
 * @category CodeIgniter
 * @license /_licenses/untermStrich/agb_normal.pdf
 * @author Christian Koller
 * @copyright 2009 untermStrich
 * @package untermStrich_database
 * @filesource
*/

if (!defined('BASEPATH')) exit('No direct script access allowed');


require_once(BASEPATH.'database/DB_active_rec.php');

/**
 * Database class for untermStrich
 *
 * @author Christian Koller
 * @package untermStrich_database
 */
class CI_DB_active_record_us extends CI_DB_active_record
{
    private $_us_cache_area = 'main';

    /**
     * The tables used within a join
     * @var        array
     */
    protected $ar_join_tables = array();

    /**
     * Commands which shall not be escaped
     */
    protected $well_known_commands = array(
        'CURRENT_TIMESTAMP',
        'NOW()+INTERVAL 1 SECOND',
        'MATCH',
        'SOUNDEX',
    );

    /**
     * Set the _protect_identifiers
     *
     * @param   boolean $escape     Escape
     */
    public function set_protect_identifiers($escape)
    {
        $this->_protect_identifiers = $escape;
    }
    
        
    /**
     * Call a stored procedure
     * 
     * @param   string    $name                     Name of procedure
     * @param   array     $args                     Array of arguments
     * @param   boolean   $use_simple_query         Use simple query
     * @param   string    &$validate                Optional validation flag for PROCEDURES which which supports validation
     * @return  mixed
     */
    public function call_procedure($name, array $args, $use_simple_query=false, &$validate=false)
    {
        foreach ($args as &$val) {
            $val = $this->escape($val);
        }

        if ($validate === true) {
            $validate = '@a'.us_uniqid('', true);
            $args[] = $validate;
        }

        $sql = 'CALL '.$name.'('.implode(', ',$args).');';
        if ($use_simple_query === true) {
            return $this->simple_query($sql);
        } else {
            return $this->query($sql);
        }
    }

    /**
     * Do a CALL procedure validation
     * 
     * @param   sting       $validate_id        The validation ID
     * @param   string      $expected_value     The expected value to compare
     * @return  mixed       Either an error message or false   
     */
    public function validate_call_procedure($validate_id, $expected_value="true")
    {
        if (!is_string($validate_id)) {
            return 'Invalid $validate_id';
        }

        $query = $this->query('SELECT '.$validate_id);
        $result = $query->row();
        $query->free_result();

        if ($result->$validate_id === $expected_value) {
            return false;
        } else {
            return 'CALL validation failed: "'.$expected_value.'" expected but got "'.$result->$validate_id.'"';
        }
    }
    
    /**
     * Where fulltext MATCH
     * 
     * @param   string      $table      Table name
     * @param   array       $fields     Field names to match
     * @param   string      $fulltext   Fulltext to search for
     * @param   boolean     $return     Optional If true the function returns the SQL, instead of setting it
     * @return  object
     */
    public function where_match($table, array $fields, $fulltext, $return=false)
    {
        $match = null;
        foreach ($fields as $field) {
            if ($match===null) {
                $match = '';
            } else {
                $match .= ', ';
            }
            $match .= $table.'.`'.$field.'`';
        }
        
        $sql = 'MATCH('.$match.') AGAINST(\''.$fulltext.'\' IN BOOLEAN MODE)';
        
        if ($return!==false) {
            return $sql;
        } else {
            return $this->where($sql);
        }
        
    }
    
    /**
     * Where fulltext MATCH using request_vars
     * 
     * @param   string      $table          Table name
     * @param   array       $fields         Field names to match
     * @param   array       $request_vars   The values to determine the resulting fulltext search string.
     * @param   boolean     $return         Optional If true the function returns the SQL, instead of setting it
     * @return  object
     */
    public function where_match_request_vars($table, array $fields, array $request_vars, $return=false)
    {
        return $this->where_match($table, $fields, $this->get_match_string($request_vars), $return);
    }
    
    /**
     * Converts the given request vars into a fulltext search string.
     * 
     * @param   array   $request_vars   The values to determine the resulting fulltext search string.
     * @return  string  Returns the fulltext search string that can be used in db query
     */
    public function get_match_string(array $request_vars)
    {
        if(count($request_vars) <= 0) {
            return "";
        }
        
        $fulltext_array = is_array($request_vars["fulltext"]) ? $request_vars["fulltext"] : array($request_vars["fulltext"]);
        $type_array = is_array($request_vars["type"]) ? $request_vars["type"] : array($request_vars["type"]);
        if (isset($request_vars["selection"])) {
            $selection_array = is_array($request_vars["selection"]) ? $request_vars["selection"] : array($request_vars["selection"]);
        } else {
            $selection_array = null;
        }
        
        $string_array = array();
        foreach($fulltext_array as $key => $value) {
            $fulltext_str = "";
            if($value) {
                if (isset($selection_array[$key]) && $selection_array[$key] == "not") {
                    $fulltext_str .= "-";
                } elseif (isset($selection_array[$key]) && $selection_array[$key] == "or") {
                    $fulltext_str .= "";
                } else {
                    $fulltext_str .= "+";
                }
                
                $search_value = trim(preg_replace('/[+\-><\(\)~*\"@]+/', ' ', $value));

                if((count($type_array) > 0) && $type_array[$key] == "exact") {
                    $fulltext_str .= "\"" . $search_value . "\"";
                } else {
                    $fulltext_str .= "*" . $search_value . "*";
                }

                array_push($string_array, $fulltext_str);
            }
        }

        return implode(" ", $string_array);
    }

    /**
     * Count all results for grouped by queries
     *
     * @param   string  $field              Alias.FieldName to group by
     * @return  integer Number of rows
     */
    public function count_all_results_grouped($field='')
    {
        $sql = $this->_compile_select('SELECT count(DISTINCT '.$field.') AS '. $this->_protect_identifiers('numrows'));

        $query = $this->query($sql);
        /*#*/#var_dump($this->last_query()); echo '<hr>';
        $this->_reset_select();

        if ($query===false) {
            return 0; //Fail
        } elseif ($query->num_rows() == 0)
        {
            $query->free_result();
            
            return 0;
        } else {
            $row = $query->row();
            $query->free_result();
            
            return intval($row->numrows);
        }
    }


    /**
     * Count all results for grouped by queries with HAVING
     *
     * @param   string  $field              Alias.FieldName to group by
     * @return  integer Number of rows
     */
    public function count_all_results_grouped_having($field='')
    {
        $sql = 'SELECT  COUNT(*) AS '. $this->_protect_identifiers('numrows').' FROM ( '.$this->_compile_select().' ) AS countAllResultsGroupedHaving';
        
        $query = $this->query($sql);
        $this->_reset_select();

        if ($query===false) {
            return 0; //Fail
        } elseif ($query->num_rows() == 0)
        {
            $query->free_result();
            
            return 0;
        } else {
            $row = $query->row();
            $query->free_result();
            
            return intval($row->numrows);
        }
    }

    /**
     * Generates the JOIN portion of the query
     *
     * @param   string  $table  The table name
     * @param   string  $cond   The join condition
     * @param   string  $type   Optional The type of join
     * @return  object
     */
    public function join($table, $cond, $type='')
    {
        $this->ar_join_tables[] = $table;
        return parent::join($table, $cond, $type);
    }

    /**
     * Generates the SELECT portion of the query to concat some fields or strings
     *
     * @param   array   $select         The fields or strings to use
     * @param   string  $name_concat    Name to use for the concatenated field
     * @return  object
     */
    public function concat(array $select, $name_concat)
    {
        $request = 'CONCAT('.implode(', ', $select).') as '.$name_concat;
        return $this->select($request, false);
    }

    /**
     * Generates the SELECT portion of the query to concat_ws some fields or strings
     *
     * @param   array   $select         The fields or strings to use
     * @param   string  $name_concat    Name to use for the concatenated field
     * @param   string  $sep            Separator
     * @param   booean  $nullif         Use nuffif or not
     * @return  object
     */
    public function concat_ws(array $select, $name_concat, $sep=' ',$nullif=true)
    {
        $select_str = '"'.$sep.'"';
        foreach ($select as $item) {
            if($nullif){
                $select_str .= ', NULLIF('.$item.', "")';
            } else {
                $select_str .= ', '.$item;
            }
        }
        $request = 'CONCAT_WS('.$select_str.') as '.$name_concat;
        return $this->select($request, false);
    }

    /**
     * Smart Escape String
     *
     * Escapes data based on type
     * Sets boolean and null types
     * Supports a defined set of SQL commands, which will not be escaped: $_well_known_commands
     *
     * @param   string  $str        String to escape
     * @param   boolean $for_blob   Default Escape for BLOB
     * @return  string  Escaped string
     */
    public function escape($str, $for_blob=false)
    {
        if (is_string($str)) {
            if (in_array($str, $this->well_known_commands)) {
                return $str;
            } else {
                if ($for_blob) {
                    return parent::escape( $str );
                } else {
                    return parent::escape( preg_replace('/\r?\n/', "\r\n", $str) );
                }
            }
        } else {
            return parent::escape($str);
        }
    }
    
    /**
     * Escape LIKE String
     *
     * Calls the individual driver for platform
     * specific escaping for LIKE conditions
     *
     * @access	public
     * @param	string
     * @return	mixed
     */
    function escape_like_str($str)
    {
        $str = $this->escape_str($str, TRUE);
        $str = str_replace('\\\\', '\\\\\\\\', $str);
        
        return $str;
    }

    /**
     * Escape the SQL Identifiers
     *
     * This function escapes column and table names
     * Supports a defined set of SQL commands, which will not be escaped: $_well_known_commands
     *
     * @param   string  $item                   The item
     * @param   boolean $prefix_single          Optional Do we prefix an item with no segments?
     * @param   boolean $protect_identifiers    Optional Call _escape_identifiers
     * @param   boolean $field_exists           Optional This flag is set when the supplied $item does not contain a field name.
     * @return  string
     *
     * @access  protected
     */
    public function _protect_identifiers($item, $prefix_single=false, $protect_identifiers=null, $field_exists=true)
    {
        if (is_string($item)) {
            //Get the first part of the item
            $item_first = $item;
            $item_a = explode(' ', $item);
            if ( count($item_a)>1 ) {
                $item_first = $item_a[0];
            }

            //Do we have a match on of the known = allowed SQL commands?
            if (in_array($item_first, $this->well_known_commands)) {
                return $item;
            } else {
                return parent::_protect_identifiers($item, $prefix_single, $protect_identifiers, $field_exists);
            }

        } else {
            return parent::_protect_identifiers($item, $prefix_single, $protect_identifiers, $field_exists);
        }
    }

    /**
     * Generates the WHERE portion of the query.
     *
     * Separates multiple calls with AND NOT
     *
     * @param   string  $key    The key to look for
     * @param   string  $value  Search value
     * @param   boolean $escape Optional Prevent data from being escaped if set to false
     * @return  object
     */
    public function not_where($key, $value=null, $escape=true)
    {
            if($value === null) {
                $key = 'NOT('.$key.')';
            }        
            return $this->_where($key, $value, 'AND NOT ', $escape);
    }

    /**
     * Generates the WHERE portion of the query.
     *
     * Separates multiple calls with OR NOT
     *
     * @param   string  $key    The key to look for
     * @param   string  $value  Search value
     * @param   boolean $escape Optional Prevent data from being escaped if set to false
     * @return  object
     */
    public function or_not_where($key, $value=null, $escape=true)
    {
            if($value === null) {
                $key = 'NOT('.$key.')';
            }
            return $this->_where($key, $value, 'OR NOT ', $escape);
    }

    /**
     * Simplifies the input for a match value
     *
     * @param   string  $value  The match value
     * @return  string  The simplified match value
     */
    public function simplify_match_value($value)
    {
        $expls = explode(' ', $value);
        $reply = '';
        foreach ($expls as $expl) {
            $expl = trim($expl);
            if (empty($expl)) continue; //Ignore empty
            
            $first = substr($expl, 0,1);
            if ($first==='+' || $first==='-' || substr($expl,-1)==='*') {
                return $value; //If any joker, no replace
            }
            $reply .= '+'.$expl.'* ';
        }
        return $reply;
    }

    /**
     * Adds an update to change_date and changed_by_id
     *
     * @param   integer $staff_id   Staff ID
     * @param   mixed   &$data      Optional Reference If set to an array,
     *                              add the info to that array, instead of to this
     * @return  void
     */
    public function add_staff_stamp($staff_id, &$data=false)
    {
        if (is_array($data)) {
            $data['change_date'] = 'CURRENT_TIMESTAMP';
            $data['changed_by_id'] = intval($staff_id);
        } else {
            $this->set('change_date', 'CURRENT_TIMESTAMP', false);
            $this->set('changed_by_id', intval($staff_id));
        }
    }

    /**
     * Adds an update to created_date and created_by_id
     *
     * @param   integer $staff_id   Staff ID
     * @param   mixed   &$data      Optional Reference If set to an array,
     *                              add the info to that array, instead of to this
     * @return  void
     */
    public function add_create_staff_stamp($staff_id, &$data=false)
    {
        if (is_array($data)) {
            $data['created_date'] = 'CURRENT_TIMESTAMP';
            $data['created_by_id'] = intval($staff_id);
        } else {
            $this->set('created_date', 'CURRENT_TIMESTAMP', false);
            $this->set('created_by_id', intval($staff_id));
        }
    }

    /**
     * Compiles an insert or update string and runs the query
     * 
     * If a row with the merge values exist it will update, else insert
     *
     * <code>Note: All values passed to $set are escaped automatically.</code>
     *
     * @param   string  $table      The table to save to
     * @param   array   $merge      An associative array of merge values
     * @param   array   $set        An associative array of insert values
     * @param   bolean  $for_temp   Flag if table must be cloned for temptable only
     * @param   string  $temptable  A copy of the original temptable
     * @param   integer &$inserted  Reference Inserted rows
     * @return  mixed
     */
    public function insert_or_update($table, array $merge, array $set, $for_temp=false, $temptable='', &$inserted=0)
    {
        //We do not want previous methodes to overwrite any defaults
        $this->_reset_write();
        $inserted = 0;
        
        //Try an update
        foreach ($merge as $key=>$value) {
            //Define what to update
            $this->where($table.'.'.$key, $value);
        }
        $ok = $this->update($table, $set);
        if (!$ok) {
            return false;
        }
        
        if ($this->affected_rows()<1) {
            $inserted = 1;
            
            //We have to write it (INSERT IF NOT EXITS)
            return $this->insert_if_not_exists($table, $merge, $set, $for_temp, $temptable);
            
//            //We have to write it
//            $set_both = array_merge($set, $merge);
//            return $this->insert($table, $set_both);            
        }
        
        return true;
    }
    
    /**
     * Compiles an insert if not existst string and runs the query
     * 
     * If a row with the merge values not exist it will insert
     *
     * <code>Note: All values passed to $set are escaped automatically.</code>
     *
     * @param   string  $table      The table to save to
     * @param   array   $merge      An associative array of merge values
     * @param   array   $set        An associative array of insert values
     * @param   bolean  $for_temp   Flag if table must be cloned for temptable only
     * @param   string  $temptable  A copy of the original temptable
     * @return  mixed
     */
    public function insert_if_not_exists($table, array $merge, array $set, $for_temp=false, $temptable='')
    {
        //We have to write it (INSERT IF NOT EXITS)

        //EXAMPLE:
        // 
        //INSERT INTO zusatz_personen(Personen_LfdNr, web)
        //SELECT 300, 'www'
        //FROM mutex
        //    LEFT OUTER JOIN zusatz_personen AS t2
        //    ON t2.Personen_LfdNr = 300
        //WHERE mutex.i=1 AND t2.Personen_LfdNr is null;
        //
        //This script only inserts new rows
        
        $t2_where = '1=1 ';
        $mutex_where = 'u10_mutex.i=1 ';
        foreach ($merge as $key=>$value) {
            //Add the merge values for insert
            $set[$key] = $value;
            $t2_where .= ' AND '.$this->_protect_identifiers('t2.'.$key).'='.$this->escape($value);
            $mutex_where .= ' AND '.$this->_protect_identifiers('t2.'.$key).' IS NULL ';
        }

        //Prepare into and select
        $into = array();
        $select = array();
        foreach ($set as $key=>$value) {
            array_push($into, $this->_protect_identifiers($key));
            array_push($select, $this->escape($value));
        }

        $sql = sprintf(
            "INSERT INTO %s (%s) SELECT %s FROM u10_mutex LEFT OUTER JOIN %s AS t2 ON %s WHERE %s",
            $this->_protect_identifiers($table),
            implode(', ', $into),
            implode(', ', $select),
            ($for_temp) ? $this->_protect_identifiers($temptable) : $this->_protect_identifiers($table),
            $t2_where,
            $mutex_where
        );

        return $this->query($sql);        
    }

    /**
     * Simple Subquery
     *
     * <code>$field IN (SELECT $group_func($table.$select) FROM $table WHERE $table.$match=$match_val)</code>
     *
     * @param   string  $field      Field name to match (including alias/table)
     * @param   string  $table      Subquery's table name
     * @param   string  $select     Subquery's field name to select
     * @param   string  $match      Subquery's field name to match
     * @param   mixed   $match_val  Optional Subquery's value to search for
     * @param   string  $group_func Optional A Group by function for the subquery field
     * @param   boolean $use_or     Use OR to add
     * @param   boolean $use_not    Use NOT... to add
     * @return  object
     */
    public function subquery_by_val($field, $table, $select, $match, $match_val=null, $group_func='', $use_or=false, $use_not=false)
    {
        if ($match_val===null) {
            $match_val = '';
        } else {
            $match_val = '='.$this->escape($match_val);
        }
        
        $sql = sprintf(
                '%s %s IN (SELECT %s(%s.%s) FROM %s WHERE %s.%s %s)',
                $field,
                ($use_not) ? 'NOT' : '',
                $group_func,
                $this->_protect_identifiers($table), $this->_protect_identifiers($select),
                $this->_protect_identifiers($table),
                $this->_protect_identifiers($table), $this->_protect_identifiers($match),
                $match_val
            );
        
        if ($use_or) {
            return $this->or_where($sql);
        } else {
            return $this->where($sql);
        }
    }
    
    /**
     * Simple Subquery with array
     *
     * <code>$field IN (SELECT $table.$select FROM $table WHERE $table.$match1=$match1_val OR  $table.$match2=$match2_val)</code>
     *
     * @param   string  $field          Field name to match (including alias/table)
     * @param   string  $table          Subquery's table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $match1         Subquery's field name to match
     * @param   array   $match1_values  Subquery's values to search for
     * @param   string  $match2         Subquery's field name to match
     * @param   array   $match2_values  Subquery's values to search for
     * @return  object
     */
    public function subquery_by_values($field, $table, $select, $match1, $match1_values)
    {
        $sql = sprintf(
                '%s IN (SELECT %s.%s FROM %s WHERE %s.%s IN(%s))',
                $field,
                $this->_protect_identifiers($table), $this->_protect_identifiers($select),
                $this->_protect_identifiers($table),
                $this->_protect_identifiers($table), $this->_protect_identifiers($match1),
                implode(', ', $match1_values)
            );
        
        return $this->where($sql);
    }
    
    /**
     * Simple Subquery for two matches with array
     *
     * <code>$field IN (SELECT $table.$select FROM $table WHERE $table.$match1=$match1_val OR  $table.$match2=$match2_val)</code>
     *
     * @param   string  $field          Field name to match (including alias/table)
     * @param   string  $table          Subquery's table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $match1         Subquery's field name to match
     * @param   array   $match1_values  Subquery's values to search for
     * @param   string  $match2         Subquery's field name to match
     * @param   array   $match2_values  Subquery's values to search for
     * @return  object
     */
    public function subquery_by_or_values($field, $table, $select, $match1, $match1_values, $match2, $match2_values)
    {
        $sql = sprintf(
                '%s IN (SELECT %s.%s FROM %s WHERE %s.%s IN(%s) OR %s.%s IN(%s))',
                $field,
                $this->_protect_identifiers($table), $this->_protect_identifiers($select),
                $this->_protect_identifiers($table),
                $this->_protect_identifiers($table), $this->_protect_identifiers($match1),
                implode(', ', $match1_values),
                $this->_protect_identifiers($table), $this->_protect_identifiers($match2),
                implode(', ', $match2_values)
            );
        
        return $this->where($sql);
    }
    
    /**
     * Create the sql for select_subquery and subquery_by_main
     *
     * @param   string  $main_table     Table name (or alias)
     * @param   string  $table          Subquery's table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $match          Subquery's field name to match
     * @param   mixed   $match_field    Subquery's match field to search for
     * @param   string  $group_func     Optional A Group by function for the subquery field
     * @param   array   $match_values   Optional An additional array of values to match (Will be prefixed with sub alias, if the value is not null and the string does not contain %SUB_ALIAS)
     * @param   string  $group_func2    Optional A 2nd Group by function for the subquery field
     * @param   mixed   $order_by_limit Optional Add order by, limit for the id defined
     * @param   mixed   &$sub_alias     Optional Reference sub_alias
     * @return  string 
     */
    protected function _subquery_by_main_create($main_table, $table, $select, $match, $match_field, $group_func='', array $match_values=array(), $group_func2='', $order_by_limit=false, &$sub_alias=false)
    {
        $sub_alias = substr($table, 0, 5).us_uniqid();

        $sql_match_values = '';
        foreach ($match_values as $key => $value) {
            if ($value!==false) {
                $sql_match_values .= ' AND '.$sub_alias.'.'.$key.'=\''.$value.'\'';
            } elseif (strpos($key, '%SUB_ALIAS')!==false) {
                $sql_match_values .= ' AND '.str_replace('%SUB_ALIAS', $sub_alias, $key);
            } else {
                $sql_match_values .= ' AND '.$sub_alias.'.'.$key;
            }
        }
        
        $sql_order_by_limit = '';
        if ($order_by_limit!==false) {
            $sql_order_by_limit = 'ORDER BY '.$sub_alias.'.'.$order_by_limit.' DESC LIMIT 1';
        }
        
        //Support DISTINCT for group by
        $group_prefix = '';
        if (stripos($select, 'DISTINCT ')===0) {
            $select = substr($select, 9);
            $group_prefix = 'DISTINCT ';
        }
        
        return sprintf(
            '(SELECT %s(%s%s(%s.%s)) FROM %s AS %s WHERE %s.%s=%s.%s %s %s)',
            $group_func, $group_prefix, 
            $group_func2,
            $sub_alias, $this->_protect_identifiers($select),
            $this->_protect_identifiers($table),
            $sub_alias,
            $sub_alias, $this->_protect_identifiers($match),
            $this->_protect_identifiers($main_table), $this->_protect_identifiers($match_field),
            $sql_match_values,
            $sql_order_by_limit
        );
    }


    /**
     * Select as subquery
     *
     * <code>(SELECT $group_func($sub_alias.$select) FROM $table AS $sub_alias WHERE $sub_alias.$match=$main_table.$match_field "ORDER BY $order_by_limit LIMIT 1") AS $as</code>
     *
     * @param   string  $main_table     Table name (or alias)
     * @param   string  $as             Alias for the subquery result
     * @param   string  $table          Subquery's table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $match          Subquery's field name to match
     * @param   mixed   $match_field    Subquery's match field to search for (from main table)
     * @param   string  $group_func     Optional A Group by function for the subquery field
     * @param   array   $match_values   Optional An additional array of values to match
     * @param   string  $group_func2    Optional A 2nd Group by function for the subquery field
     * @param   mixed   $order_by_limit Optional Add order by, limit for the id defined
     * @return  object
     */
    public function select_subquery($main_table, $as, $table, $select, $match, $match_field, $group_func='', array $match_values=array(), $group_func2='', $order_by_limit=false)
    {
        //Add AS ...
        $sql = sprintf(
            '%s AS %s',
            $this->_subquery_by_main_create($main_table, $table, $select, $match, $match_field, $group_func, $match_values, $group_func2, $order_by_limit),
            $as
        );
        $this->select($sql);
    }
    
    /**
     * Create select as subquery
     *
     * <code>(SELECT $group_func($sub_alias.$select) FROM $table AS $sub_alias WHERE $sub_alias.$match=$main_table.$match_field "ORDER BY $order_by_limit LIMIT 1")</code>
     *
     * @param   string  $main_table     Table name (or alias)
     * @param   string  $table          Subquery's table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $match          Subquery's field name to match
     * @param   mixed   $match_field    Subquery's match field to search for (from main table)
     * @param   string  $group_func     Optional A Group by function for the subquery field
     * @param   array   $match_values   Optional An additional array of values to match
     * @param   string  $group_func2    Optional A 2nd Group by function for the subquery field
     * @param   mixed   $order_by_limit Optional Add order by, limit for the id defined
     * @return  object
     */
    public function create_select_subquery($main_table, $table, $select, $match, $match_field, $group_func='', array $match_values=array(), $group_func2='', $order_by_limit=false)
    {
        return $this->_subquery_by_main_create($main_table, $table, $select, $match, $match_field, $group_func, $match_values, $group_func2, $order_by_limit);
    }

    /**
     * Select as subquery
     *
     * <code>(SELECT $group_func($sub_alias.$select) FROM $table AS $sub_alias WHERE $sub_alias.$match=$main_table.$match_field) AS $as</code>
     *
     * @param   string  $main_table     Table name (or alias)
     * @param   string  $as             Alias for the subquery result
     * @param   string  $table          Subquery's table name
     * @param   string  $trans_table    Translation table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $trans_select   Translation Subquery's field name to select
     * @param   string  $match          Subquery's field name to match
     * @param   string  $trans_match    Translation Subquery's field name to match
     * @param   mixed   $match_field    Subquery's match field to search for
     * @param   array   $match_values   Optional An additional array of values to match
     * @param   string  $group_func1    Optional A Group by function for the subquery field
     * @param   string  $group_func2    Optional A Part two of the group by function for the subquery field
     * @return  object
     */
    public function select_subquery_multilingual($main_table, $as, $table, $trans_table, $select, $trans_select, $match, $trans_match, $match_field, array $match_values=array(), $group_func1='', $group_func2='')
    {
        $sql = $this->create_select_subquery_multilingual($main_table, $table, $trans_table, $select, $trans_select, $match, $trans_match, $match_field, $match_values, $group_func1, $group_func2);
        
        //Add as
        $sql .=' AS '.$as;
        
        $this->select($sql);
    }
        
    /**
     * Create a select as subquery
     *
     * <code>(SELECT $group_func($sub_alias.$select) FROM $table AS $sub_alias WHERE $sub_alias.$match=$main_table.$match_field)</code>
     *
     * @param   string  $main_table     Table name (or alias)
     * @param   string  $table          Subquery's table name
     * @param   string  $trans_table    Translation table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $trans_select   Translation Subquery's field name to select
     * @param   string  $match          Subquery's field name to match
     * @param   string  $trans_match    Translation Subquery's field name to match
     * @param   mixed   $match_field    Subquery's match field to search for
     * @param   array   $match_values   Optional An additional array of values to match
     * @param   string  $group_func1    Optional A Group by function for the subquery field
     * @param   string  $group_func2    Optional A Part two of the group by function for the subquery field
     * @return  object
     */
    public function create_select_subquery_multilingual($main_table, $table, $trans_table, $select, $trans_select, $match, $trans_match, $match_field, array $match_values=array(), $group_func1='', $group_func2='')
    {
        $CI = CI_Controller::get_instance();
        
        //Get main table SQL
        $sub_alias = false;
        $sql_main = $this->_subquery_by_main_create($main_table, $table, $select, $match, $match_field, '', $match_values, '', false, $sub_alias);
        
        //Get translation table SQL
        $trans_match_values = array(
            'language'  => $CI->web_settings->getLanguage(),
        );
        $sql_trans = $this->_subquery_by_main_create($sub_alias, $trans_table, $trans_select, $trans_match, 'id', '', $trans_match_values);
        
        //Replace select field in main table with tranlsation or main table COALESCE
        $to_replace = '(('.$sub_alias.'.'.$select.'))';
        $sql_replace = sprintf(
            '%s( COALESCE(%s, %s) %s)',
            $group_func1,
            $sql_trans,
            $to_replace,
            $group_func2
        );
        $sql = str_replace($to_replace, $sql_replace, $sql_main);
        
        $to_replace = '(('.$sub_alias.'.`'.$select.'`))';
        return str_replace($to_replace, $sql_replace, $sql);
    }

    /**
     * Subquery by main field - correlated subquery
     *
     * <code>$main_table.$field IN (SELECT $group_func($sub_alias.$select) FROM $table AS $sub_alias WHERE $sub_alias.$match=$main_table.$match_field)</code>
     *
     * @param   string  $main_table     Table name (or alias)
     * @param   string  $field          Field name to match (including alias/table)
     * @param   string  $table          Subquery's table name
     * @param   string  $select         Subquery's field name to select
     * @param   string  $match          Subquery's field name to match
     * @param   mixed   $match_field    Subquery's match field to search for
     * @param   string  $group_func     Optional A Group by function for the subquery field
     * @param   array   $match_values   Optional An additional array of values to match
     * @return  object
     */
    public function subquery_by_main($main_table, $field, $table, $select, $match, $match_field, $group_func='', array $match_values=array())
    {
        //Add where ... IN
        $sql = sprintf(
            '%s.%s IN %s',
            $this->_protect_identifiers($main_table), $field,
            $this->_subquery_by_main_create($main_table, $table, $select, $match, $match_field, $group_func, $match_values)
        );
        
        return $this->where($sql);
    }

    /**
     * Simple insert into select
     *
     * <code>INSERT INTO $into_table SELECT $map...,$vars...
     *       FROM $from_table WHERE $search_match=$search_value</code>
     *
     * <b>Take care: The into table has to have the same fields as the from table for those fields defined as the first argument</b>
     *
     * @param   array   $map            Fields to copy from one to the other table (supports assoc and simple arrays)
     * @param   array   $vars           Field=>value list of fields to set manually
     * @param   string  $into_table     The table to insert the data into
     * @param   string  $from_table     The table to read the data from
     * @param   mixed   $search_match   Optional The field to match within the table to read from
     * @param   mixed   $search_value   Optional The value to search for
     * @param   mixed   $search_match2  Optional The field 2 to match within the table to read from
     * @param   mixed   $search_value2  Optional The value 2 to search for
     * @return  object
     */
    public function insert_select(array $map, array $vars, $into_table, $from_table, $search_match=false, $search_value=false, $search_match2=false, $search_value2=false)
    {
        //The mapping fields
        $into = array();
        $select = array();

        //Translate mapping
        foreach ($map as $mapkey => $mapfield) {
            if (is_numeric($mapkey)) {
                $mapkey = $mapfield;
            }
            $into[] = $this->_protect_identifiers($mapkey);
            $select[] = $this->_protect_identifiers($mapfield);
        }
        foreach ($vars as $key => $value) {
            $into[] = $this->_protect_identifiers($key);
            $select[] = $this->escape($value);
        }

        $sql = sprintf(
            "INSERT INTO %s (%s) SELECT %s FROM %s",
            $this->_protect_identifiers($into_table),
            implode(', ', $into),
            implode(', ', $select),
            $this->_protect_identifiers($from_table)
        );

        if ($search_match!==false && $search_value!==false) {
            $sql .= sprintf(
                " WHERE %s='%s'",
                $search_match,
                $search_value
            );
        }

        if ($search_match2!==false && $search_value2!==false) {
            $sql .= sprintf(
                " AND %s='%s'",
                $search_match2,
                $search_value2
            );
        }

        return $this->query($sql);
    }

    /**
     * Return 1 if true, else 0
     *
     * <code>IF($table.$field $comp \'$value\', 1, 0) AS $fieldreturn</code>
     *
     * @param   string  $table          Table name
     * @param   string  $field          Field
     * @param   string  $comp           Comparison - One of {@link http://dev.mysql.com/doc/refman/5.1/en/comparison-operators.html}
     * @param   string  $value          Value
     * @param   string  $fieldreturn    A name for the field in the result
     * @param   boolean $return         Optional If true the function returns the SQL, instead of setting it
     * @return  mixed   Either an object or the SQL string
     */
    public function true_if($table, $field, $comp, $value, $fieldreturn, $return=false)
    {
        $sql = sprintf(
            'IF(%s.%s %s \'%s\', 1, 0) AS %s',
            $this->_protect_identifiers($table), $this->_protect_identifiers($field),
            $comp,
            $value,
            $fieldreturn
        );
        if ($return!==false) {
            return $sql;
        } else {
            return $this->select($sql);
        }
    }

    /**
     * Select using a if
     *
     * <code>IF($table.$field $comp \'$value\', \'$value\', $table.$field) AS $fieldreturn</code>
     *
     * @param   string  $table          Table name
     * @param   string  $field          Field
     * @param   string  $comp           Comparison - One of {@link http://dev.mysql.com/doc/refman/5.1/en/comparison-operators.html}
     * @param   string  $value          Value
     * @param   string  $fieldreturn    Optional A name for the field in the result (Default = null) (null = No as)
     * @param   boolean $return         Optional If true the function returns the SQL, instead of setting it
     * @return  mixed   Either an object or the SQL string
     */
    public function select_if($table, $field, $comp, $value, $fieldreturn=null, $return=false)
    {
        $sql = sprintf(
            'IF(%s.%s %s \'%s\', \'%s\', %s.%s)%s',
            $this->_protect_identifiers($table), $this->_protect_identifiers($field),
            $comp,
            $value,
            $value,
            $this->_protect_identifiers($table), $this->_protect_identifiers($field),
            ($fieldreturn===null) ? '' : ' AS '.$fieldreturn
        );
        if ($return!==false) {
            return $sql;
        } else {
            return $this->select($sql);
        }
    }

    /**
     * Select NULL if the IF is true, else the field
     * 
     * @param   string  $table          Table name
     * @param   string  $field          Field to return
     * @param   string  $field_comp     Field for comparison
     * @param   string  $comp           Comparison - One of {@link http://dev.mysql.com/doc/refman/5.1/en/comparison-operators.html}
     * @param   string  $value          Value
     * @param   string  $fieldreturn    Optional A name for the field in the result (Default = null) (null = No as)
     * @param   boolean $return         Optional If true the function returns the SQL, instead of setting it
     * @return  mixed   Either an object or the SQL string
     */
    public function select_null_if($table, $field, $field_comp, $comp, $value, $fieldreturn=null, $return=false)
    {
        $sql = sprintf(
            'IF(%s.%s %s \'%s\', NULL, %s.%s)%s',
            $this->_protect_identifiers($table), $this->_protect_identifiers($field_comp),
            $comp,
            $value,
            $this->_protect_identifiers($table), $this->_protect_identifiers($field),
            ($fieldreturn===null) ? '' : ' AS '.$fieldreturn
        );
        if ($return!==false) {
            return $sql;
        } else {
            return $this->select($sql, false);
        }
    }

    /**
     * Was either the table or the table alias already used in a join?
     *
     * @param   string  $table_or_alias     Table or alias name
     * @return  boolean Joined?
     */
    public function was_joined($table_or_alias)
    {
        $join_tables = ' '.implode(' ',$this->ar_join_tables).' ';
        return (strpos($join_tables, ' '.$table_or_alias.' ')!==false);
    }

    /**
     * Display an error message
     *
     * @param   string      $error      The error message
     * @param   string      $swap       Any "swap" values
     * @param   boolean     $native     Not in use
     * @return  mixed
     */
    public function display_error($error='', $swap='', $native=false)
    {
        //Get message for the error log
        $LANGen = clone load_class('Lang');
        $LANGen->load('db', 'english');
        $logmessage = (is_array($error)) ? implode(' | ', $error).' | Swap: '.$swap : str_replace('%s', $swap, $LANGen->line($error));
        log_message('error', 'Database error: '.$logmessage);
        
        if (!is_array($error)) {
            //Only catch non query related errors a total fail
            parent::display_error($error, $swap, $native);
            exit;
        } else {
            return false;
        }
    }

    /**
     * Resets the active record values.  Called by the get() function
     *
     * @return  void
     * @access  private
     */
    public function _reset_select()
    {
        $this->ar_join_tables = array();
        parent::_reset_select();
    }

    /**
     * Creates the search string out of a value
     *
     * It adds %...% if there is no * in the string, else it replaces the * with %
     *
     * @param   string  &$value     Reference The value
     * @return  void
     */
    public function exchange_or_add_percentage(&$value)
    {
        //Escape some special characters
        $from = array('_',  '%',  '?');
        $to   = array('\_', '\%', '_');
        $value = str_replace($from, $to, $value);
        
        //It adds %...% if there is no * in the string, else it replaces the * with %
        $add_both = true;
        
        //Change at word start
        if ( substr($value, 0, 1)=='*' ) {
            $value = '%'.substr($value, 1);
            $add_both = false;
        }
        
        //Change at word end
        if ( substr($value, -1)=='*' ) {
            $value = substr($value, 0, -1).'%';
            $add_both = false;
        }
        
        //Replace others
        $value = strtr($value, '*', '%');
        
        //Add on both sites
        if ($add_both) {
            $value = '%'.$value.'%';
        }
    }

    /**
     * Enable Query Caching and if set, sets the name of the caching area
     *
     * @param   string  $cache_area     Cache area name
     * @return	void
     * @todo Datenbank cache deaktivieren ist hier moeglich
     */
    public function cache_on($cache_area='main')
    {
        $this->_us_cache_area = $cache_area;
        $this->cache_on = TRUE;
        return TRUE;
    }

    /**
     * Get the cache area
     *
     * @return  string
     */
    public function get_us_cache_area() {
        return $this->_us_cache_area;
    }
    
    /**
     * Cleanup memory of a MEMORY table
     * 
     * @param   string  $name       Table name
     * @return  mixed   Either false or an error message
     */
    public function clean_memory_table($name)
    {
        //The create part
        if (!$this->query('DROP TABLE IF EXISTS `'.$name.'_tmp`;')) {
            return 'Failed to drop temp table for '.$name;
        }
        if (!$this->query('CREATE TEMPORARY TABLE `'.$name.'_tmp` AS SELECT * FROM `'.$name.'`;')) {
            return 'Failed to create temp table for '.$name;
        }
        //Clean table
        if (!$this->query('TRUNCATE TABLE `'.$name.'`;')) {
            return 'Failed to truncate table for '.$name;
        }
        //Copy back
        if (!$this->query('INSERT INTO `'.$name.'` SELECT * FROM `'.$name.'_tmp`;')) {
            return 'Failed to recover from temp table for '.$name;
        }
        //Cleanup
        if (!$this->query('DROP TABLE IF EXISTS `'.$name.'_tmp`;')) {
            return 'Failed to drop temp table for '.$name;
        }
        return false;
    }
    
	/**
	 * Get
	 *
	 * Compiles the select statement based on the other functions called
	 * and runs the query
	 *
	 * @access	public
	 * @param	string	the table
	 * @param	string	the limit clause
	 * @param	string	the offset clause
	 * @return	object
	 */
	function get($table = '', $limit = null, $offset = null)
	{
		if ($table != '')
		{
			$this->_track_aliases($table);
			$this->from($table);
		}

		if ( ! is_null($limit))
		{
			$this->limit($limit, $offset);
		}
        #####################################################################################################################################################
        ### CW: Try to ensure to also order_by(id) if we use limit => we have to ensure a deterministic sort order (as of MySQL 8) for correct pagination ###
        ### For details see: https://blog.kalvad.com/sql-paging-requires-a-deterministic-sort-order-a-classic-example-of-s-e-p-somebody-elses-problem/ ######
        #####################################################################################################################################################
        if (!empty($this->ar_limit)) {
            $order_added = false;
            //Get alias from main table => prefer id of main table
            $table_alias = (isset($this->ar_select[0]) && stripos($this->ar_select[0], ' AS ') !== false) ? trim(substr($this->ar_select[0], strrpos($this->ar_select[0], ' '))) : false;
            if ($table_alias) {
                foreach ($this->ar_select as $field) {
                    if (strtolower($field) === "$table_alias.id") {
                        $this->order_by($field, 'asc');
                        $order_added = true;
                        break;
                    }
                }
            }
            //If main table does not have an alias or alias.id field is not selected => test for other id field names
            if ($order_added === false) {
                foreach ($this->ar_select as $field){
                    if (strpos($field, '(') === false && (strtolower($field) === 'id' || substr(strtolower($field), -3) === '.id')) {
                        $this->order_by($field, 'asc');
                        break;
                    }
                }
            }
        }
        #####################################################################################################################################################
        #####################################################################################################################################################

		$sql = $this->_compile_select();
		$result = $this->query($sql);
		$this->_reset_select();
		return $result;
	}    
    
} // END class CI_DB_active_record_us